# -*- coding: utf-8 -*-

"""
This module contains usefull functions commons to every modules.
"""

import os
import logging
from itertools import zip_longest
from datetime import datetime
from typing import Iterable
import re

from .config import Config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# useful dates formats encountered when working with Sentinel-2 data
DATES_FORMATS = dict(
    ymd="%Y-%m-%d",
    filename="%Y%m%dT%H%M%S",
    metadata="%Y-%m-%dT%H:%M:%S.%fZ")


def format_word(word: str) -> str:
    """
    Formats a word to lower case and removes heading and trailing spaces.

    :param word: a string to format.
    """
    if not isinstance(word, str):
        raise TypeError("word has to be a string")
    return word.lower().strip()


def grouper(iterable, n: int, fillvalue: bool = None) -> Iterable:
    """
    Collects data into fixed-length chunks or blocks.

    :param iterable: an iterable.
    :param n: number of items per chunk/block.
    :param fillvalue: replacing value for missing items in the chunk.
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def str_to_datetime(date_string: str, date_format: str) -> datetime:
    """
    Returns a datetime object from a string according to a format.

    :param date_string: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strptime(date_string, date_format)


def datetime_to_str(date: datetime, date_format: str) -> str:
    """
    Returns a string from a datetime object according to a format.

    :param date: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strftime(date, date_format)


def human_size(bytes, units=['B','KB','MB','GB','TB', 'PB', 'EB']):
    """ Returns a human readable string reprentation of bytes"""
    return str(bytes) + units[0] if bytes < 1024.0 else human_size(bytes>>10, units[1:])
    
def human_size_decimal(size, decimal_places=2):
    for unit in ['B','KB','MB','GB','TB', 'PB', 'EB']:
        if size < 1024.0:
            break
        size /= 1024.0
    return "{}{}".format(format(size, '.' + str(decimal_places) + 'f'), unit)
    #~ return f"{size:.{decimal_places}f}{unit}"

def getFolderSize(folder, follow_symlinks = False):
    total_size = os.path.getsize(folder)
    try:
        for item in os.listdir(folder):
            itempath = os.path.join(folder, item)
            if os.path.isfile(itempath) and (follow_symlinks or not os.path.islink(itempath)):
                total_size += os.path.getsize(itempath)
            elif os.path.isdir(itempath) and (follow_symlinks or not os.path.islink(itempath)):
                total_size += getFolderSize(itempath)
    except:
        pass
    return total_size
    
def setPermissions(path):
    for dir_path, dir_names, files in os.walk(str(path)):
        for d in dir_names:
            name = dir_path + '/' + d
            os.chmod(name, os.stat(name).st_mode | 0o075)
            os.chmod(name, os.stat(name).st_mode & ~0o002)
            
        for f in files:
            name = dir_path + '/' + f
            os.chmod(name, os.stat(name).st_mode | 0o064)
            os.chmod(name, os.stat(name).st_mode & ~0o003)
            
def get_current_Sen2Cor_version():
    """ Returns your current Sen2Cor version """
    sen2cor_bashrc_path = Config().get("sen2cor_bashrc_path")
    return next(iter(re.findall('Sen2Cor-(\d{2}\.\d{2}\.\d{2})', str(sen2cor_bashrc_path))), None)


    

